"""
Static data collection.
"""

from .parse import parse_Pct_file_name as parse_Pct_file_name
from .parse import parse_Pct_two_files as parse_Pct_two_files

from .static_data import FARM as FARM
from .static_data import STATES as STATES
from .static_data import PCTCURVE as PCTCURVE
from .static_data import StaticData as StaticData
