"""
Functions for adding turbines to the wind farm.
"""

from .row import add_row as add_row
from .grid import add_grid as add_grid
from .from_json import add_from_json as add_from_json
from .from_csv import add_from_csv as add_from_csv
from .from_eww import add_from_eww as add_from_eww
from .from_wrf import add_from_wrf as add_from_wrf
from .from_file import add_from_file as add_from_file
from .from_df import add_from_df as add_from_df
from .from_random import add_random as add_random
from .ring import add_ring as add_ring
from .from_arrays import add_from_arrays as add_from_arrays
