"""
Geometries in two dimensions.
"""

from .area_geometry import AreaGeometry as AreaGeometry
from .area_geometry import InvertedAreaGeometry as InvertedAreaGeometry
from .area_geometry import AreaUnion as AreaUnion
from .area_geometry import AreaIntersection as AreaIntersection
from .polygon import ClosedPolygon as ClosedPolygon
from .circle import Circle as Circle
from .half_plane import HalfPlane as HalfPlane
