"""
Wind deficit wake models.
"""

from .jensen import JensenWake as JensenWake
from .bastankhah14 import Bastankhah2014 as Bastankhah2014
from .bastankhah16 import Bastankhah2016Model as Bastankhah2016Model
from .bastankhah16 import Bastankhah2016 as Bastankhah2016
from .turbopark import TurbOParkWake as TurbOParkWake
from .turbopark import TurbOParkWakeIX as TurbOParkWakeIX
